"""
Copyright (C) 2022  ETH Zurich, Manuel Kaufmann, Velko Vechev

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import moderngl
import numpy as np
import trimesh

from aitviewer.renderables.meshes import Meshes
from aitviewer.scene.material import Material
from aitviewer.scene.node import Node
from aitviewer.shaders import get_cylinder_program
from aitviewer.utils import set_lights_in_program
from aitviewer.utils import set_material_properties
from aitviewer.utils import compute_vertex_and_face_normals
from aitviewer.utils.so3 import aa2rot_numpy as aa2rot
from moderngl_window.opengl.vao import VAO


_CYLINDER_SECTORS = 8


def _create_disk(n_disks=1, radius=1.0, sectors=None, plane='xz'):
    """
    Create `n_disks` many disks centered at the origin with the given radius1.
    :param n_disks: How many disks to create.
    :param radius: Radius of the disks.
    :param sectors: How many lines to use to approximate the disk. Increasing this will lead to more vertex data.
    :param plane: In which plane to create the disk.
    :return: Vertices as a np array of shape (N, V, 3), and face data as a np array of shape (F, 3).
    """
    assert plane in ['xz', 'xy', 'yz']
    sectors = sectors or _CYLINDER_SECTORS
    angle = 2 * np.pi / sectors

    c1 = 'xyz'.index(plane[0])
    c2 = 'xyz'.index(plane[1])
    c3 = 'xyz'.index('xyz'.replace(plane[0], '').replace(plane[1], ''))

    # Vertex Data.
    vertices = np.zeros((n_disks, sectors + 1, 3))
    x = radius * np.cos(np.arange(sectors) * angle)
    y = radius * np.sin(np.arange(sectors) * angle)
    vertices[:, 1:, c1] = x
    vertices[:, 1:, c2] = y

    # Faces.
    faces = np.zeros((sectors, 3), dtype=np.int32)
    idxs = np.array(range(1, sectors+1), dtype=np.int32)
    faces[:, 2] = idxs
    faces[:-1, 1] = idxs[1:]
    faces[-1, 1] = 1

    return {'vertices': vertices, 'faces': faces}


def _create_cylinder_from_to(v1, v2, radius1=1.0, radius2=1.0, sectors=None):
    """
    Create cylinders from points v1 to v2.
    :param v1: A np array of shape (N, 3).
    :param v2: A np array of shape (N, 3).
    :param radius1: The radius at the bottom of the cylinder.
    :param radius2: The radius at the top of the cylinder.
    :param sectors: how many lines to use to approximate the bottom disk.
    :return: Vertices and normals as a np array of shape (N, V, 3) and face data in shape (F, 3), i.e. only one
      face array is created for all cylinders.
    """
    sectors = sectors or _CYLINDER_SECTORS
    n_cylinders = v1.shape[0]

    # Create bottom lid.
    bottom = _create_disk(n_disks=n_cylinders, radius=radius1, sectors=sectors)

    # We must also change the winding of the bottom triangles because we have backface culling enabled and
    # otherwise we wouldn't see the bottom lid even if the normals are correct.
    fs_bottom = bottom['faces']
    fs_bottom[:, 1], fs_bottom[:, 2] = fs_bottom[:, 2], fs_bottom[:, 1].copy()

    # Create top lid.
    top = _create_disk(n_disks=n_cylinders, radius=radius2, sectors=sectors)
    p2 = np.zeros((n_cylinders, 3))
    p2[:, 1] = np.linalg.norm(v2 - v1, axis=-1)
    top['vertices'] = top['vertices'] + p2[:, np.newaxis]

    # Shift indices of top faces by how many vertices the bottom lid has.
    n_vertices = bottom['vertices'].shape[1]
    fs_top = top['faces'] + n_vertices

    # Create the faces that make up the coat between bottom and top lid.
    idxs_bot = np.array(range(1, sectors+1), dtype=np.int32)
    idxs_top = idxs_bot + n_vertices
    fs_coat1 = np.zeros((sectors, 3), dtype=np.int32)
    fs_coat1[:, 0] = idxs_top
    fs_coat1[:-1, 1] = idxs_top[1:]
    fs_coat1[-1, 1] = idxs_top[0]
    fs_coat1[:, 2] = idxs_bot

    fs_coat2 = np.zeros((sectors, 3), dtype=np.int32)
    fs_coat2[:, 0] = fs_coat1[:, 1]
    fs_coat2[:-1, 1] = idxs_bot[1:]
    fs_coat2[-1, 1] = idxs_bot[0]
    fs_coat2[:, 2] = idxs_bot

    # Concatenate everything to create a single mesh.
    vs = np.concatenate([bottom['vertices'], top['vertices']], axis=1)
    fs = np.concatenate([fs_bottom, fs_top, fs_coat1, fs_coat2], axis=0)

    # Compute smooth normals.
    vertex_faces = trimesh.Trimesh(vs[0], fs, process=False).vertex_faces
    ns, _ = compute_vertex_and_face_normals(vs[0:1], fs, vertex_faces, normalize=True)
    ns = np.repeat(ns, n_cylinders, axis=0)

    # Rotate cylinders to align the the given data.
    vs, ns = _rotate_cylinder_to(v2-v1, vs, ns)

    # Translate cylinders to the given positions
    vs += v1[:, np.newaxis]

    return {'vertices': vs, 'normals': ns, 'faces': fs}


def _create_cone_from_to(v1, v2, radius=1.0, sectors=None):
    """
    Create a cone from points v1 to v2.
    :param v1: A np array of shape (N, 3).
    :param v2: A np array of shape (N, 3).
    :param radius: The radius for the disk of the cone.
    :param sectors: how many lines to use to approximate the bottom disk.
    :return: Vertices and normals as a np array of shape (N, V, 3) and face data in shape (F, 3), i.e. only one
      face array is created for all cones.
    """
    sectors = sectors or _CYLINDER_SECTORS
    n_cylinders = v1.shape[0]

    # Create bottom lid.
    bottom = _create_disk(n_cylinders, radius, sectors)

    # We must also change the winding of the bottom triangles because we have backface culling enabled and
    # otherwise we wouldn't see the bottom lid even if the normals are correct.
    fs_bottom = bottom['faces']
    fs_bottom[:, 1], fs_bottom[:, 2] = fs_bottom[:, 2], fs_bottom[:, 1].copy()

    # Add the top position as a new vertex.
    p2 = np.zeros((n_cylinders, 3))
    p2[:, 1] = np.linalg.norm(v2 - v1, axis=-1)
    vs = np.concatenate([bottom['vertices'], p2[:, np.newaxis]], axis=1)
    n_vertices = vs.shape[1]
    idx_top = n_vertices-1

    # Create the faces from the bottom lid to the top point.
    idxs_bot = np.array(range(1, sectors + 1), dtype=np.int32)
    fs_coat = np.zeros((sectors, 3), dtype=np.int32)
    fs_coat[:, 0] = idx_top
    fs_coat[:-1, 1] = idxs_bot[1:]
    fs_coat[-1, 1] = idxs_bot[0]
    fs_coat[:, 2] = idxs_bot

    # Concatenate everything to create a single mesh.
    fs = np.concatenate([fs_bottom, fs_coat], axis=0)

    # Compute smooth normals.
    vertex_faces = trimesh.Trimesh(vs[0], fs, process=False).vertex_faces
    ns, _ = compute_vertex_and_face_normals(vs[0:1], fs, vertex_faces, normalize=True)
    ns = np.repeat(ns, n_cylinders, axis=0)

    # Rotate cones to align the the given data.
    vs, ns = _rotate_cylinder_to(v2 - v1, vs, ns)

    # Translate cones to the given positions
    vs += v1[:, np.newaxis]

    return {'vertices': vs, 'normals': ns, 'faces': fs}


def _rotate_cylinder_to(target, vs, ns):
    """
    Rotate vertices and normals such that the main axis of the cylinder aligns with `target`.
    :param target: A np array of shape (N, 3) specifying the target direction for each cylinder given in `vs`.
    :param vs: A np array of shape (N, V, 3), i.e. the vertex data for each cylinder.
    :param ns: A np array of shape (N, V, 3), i.e. the normal data for each cylinder.
    :return: The rotated vertices and normals.
    """
    n_cylinders = vs.shape[0]
    s = np.array([[0.0, 1.0, 0.0]]).repeat(n_cylinders, axis=0)
    v = target / (np.linalg.norm(target, axis=1, keepdims=True) + 1e-16)
    a = np.cross(s, v, axis=1)
    dot = np.sum(s * v, axis=1)
    theta = np.arccos(dot)
    is_col = np.linalg.norm(a, axis=1) < 10e-6
    a[is_col] = np.array([1.0, 0.0, 0.0])
    theta[is_col] = 0.0
    theta[np.logical_and(dot < 0.0, is_col)] = np.pi
    a = a / np.linalg.norm(a, axis=1, keepdims=True)
    rot = aa2rot(a * theta[..., np.newaxis])
    vs = np.matmul(rot[:, np.newaxis], vs[..., np.newaxis]).squeeze(-1)
    ns = np.matmul(rot[:, np.newaxis], ns[..., np.newaxis]).squeeze(-1)
    return vs, ns


class Lines(Node):
    """Render lines as cylinders or cones. Can render approx. 600k lines at 40 fps."""

    def __init__(self,
                 lines,
                 r_base=0.01,
                 r_tip=None,
                 color=(0.0, 0.0, 1.0, 1.0),
                 mode='line_strip',
                 cast_shadow=True,
                 **kwargs):
        """
        Initializer.
        :param lines: Set of 3D coordinates as a np array of shape (F, L, 3).
        :param r_base: Thickness of the line.
        :param r_tip: If set, the thickness of the line will taper from r_base to r_tip. If set to 0.0 it will create
          a proper cone.
        :param color: Color of the line (4-tuple).
        :param mode: 'lines' or 'line_strip' -> ModernGL drawing mode - LINE_STRIP oder LINES
        """
        assert len(color) == 4
        assert len(lines.shape) >= 2
        assert mode == "lines" or mode == "line_strip"
        self.mode = mode
        self._lines = None
        self.lines = lines

        super(Lines, self).__init__(n_frames=self.lines.shape[0], **kwargs)

        self.r_base = r_base
        self.r_tip = r_tip

        vs, fs, ns = self.get_mesh()
        material = kwargs.get('material', Material(color=color, ambient=0.2))
        self.mesh = Meshes(vs, fs, ns, color=color, material=material, cast_shadow=cast_shadow)
        self.mesh.position = self.position
        self.add(self.mesh, has_gui=True, show_in_hierarchy=False)

    @property
    def lines(self):
        return self._lines

    @lines.setter
    def lines(self, value):
        self._lines = value if len(value.shape) == 3 else value[np.newaxis]

    def gui(self, imgui):
        self.mesh.gui(imgui)

    def redraw(self, **kwargs):
        vs, fs, ns = self.get_mesh()
        self.mesh.vertices = vs
        self.mesh.faces = fs
        super().redraw()

    @property
    def color(self):
        return self.mesh.color

    @color.setter
    def color(self, color):
        self.mesh.color = color

    def get_mesh(self):
        # Extract pairs of lines such that line i goes from v0s[i] to v1s[i].
        if self.mode == 'lines':
            assert self.lines.shape[1] % 2 == 0
            v0s = self.lines[:, ::2]
            v1s = self.lines[:, 1::2]
        else:
            v0s = self.lines[:, :-1]
            v1s = self.lines[:, 1:]

        # Data is in the form of (F, N_LINES, V, 3), convert it to (F*N_LINES, 3)
        n_lines = v0s.shape[1]
        v0s = np.reshape(v0s, (-1, 3))
        v1s = np.reshape(v1s, (-1, 3))
        self.r_tip = self.r_base if self.r_tip is None else self.r_tip
        # If r_tip is below a certain threshold, we create a proper cone, i.e. with just a single vertex at the top.
        if self.r_tip < 10e-6:
            data = _create_cone_from_to(v0s, v1s, radius=self.r_base)
        else:
            data = _create_cylinder_from_to(v0s, v1s, radius1=self.r_base, radius2=self.r_tip)

        # Convert to (F, N_LINES*V, 3)
        n_vertices = data['vertices'].shape[1]
        vs = np.reshape(data['vertices'], [self.n_frames, -1, 3])
        ns = np.reshape(data['normals'], [self.n_frames, -1, 3])
        fs = []
        for i in range(n_lines):
            fs.append(data['faces'] + i * n_vertices)
        fs = np.concatenate(fs)

        return vs, fs, ns


class LinesWithGeometryShader(Node):
    """
    Draw a line between n points using a cylinder geometry shader.
    Two radii can be set resulting in a cylinder, tapered cylinder, or an arrow (r_tip = 0.0)
    """

    def __init__(self,
                 lines,
                 r_base=0.01,
                 r_tip=None,
                 color=(0.0, 0.0, 1.0, 1.0),
                 mode='line_strip',
                 **kwargs):
        """
        Initializer.
        :param lines: Set of 3D coordinates as a np array of shape (N, L, 3).
        :param r_base: Thickness of the line.
        :param r_tip: If set, the thickness of the line will taper from r_base to r_tip
        :param color: Color of the line (4-tuple).
        :param mode: 'lines' or 'line_strip' -> ModernGL drawing mode - LINE_STRIP oder LINES
        """
        assert len(color) == 4
        assert len(lines.shape) >= 2
        assert mode == "lines" or mode == "line_strip"
        lines = lines if len(lines.shape) == 3 else lines[np.newaxis]

        super(LinesWithGeometryShader, self).__init__(n_frames=len(lines), color=color, **kwargs)

        self.r_base = r_base
        self.r_tip = r_tip if r_tip is not None else r_base
        self.lines = lines
        self.colors = np.full((self.n_lines, 4), self.color)
        self.mode = moderngl.LINE_STRIP if mode == 'line_strip' else moderngl.LINES
        self.vao = VAO("cylinder", mode=self.mode)

    @classmethod
    def from_start_end_points(cls, v0, v1, **kwargs):
        """Create lines that connect from v0[i, j] to v1[i, j] for frame i and line j."""
        vs = np.zeros((v0.shape[0], v0.shape[1] * 2, 3))
        vs[:, ::2] = v0
        vs[:, 1::2] = v1
        return cls(vs, mode='lines', **kwargs)

    def on_frame_update(self):
        if self.is_renderable:
            self.vbo_vertices.write(self.lines_current.astype('f4').tobytes())

    def update_data(self, lines):
        self.lines = lines
        self.on_frame_update()

    @property
    def n_lines(self):
        return self.lines.shape[1]

    @property
    def lines_current(self):
        return self.lines[self.current_frame_id]

    @Node.color.setter
    def color(self, color):
        self.material.color = color
        self.colors = np.full((self.n_lines, 4), self.color)
        if self.is_renderable:
            self.vbo_colors.write(self.colors.astype('f4').tobytes())

    # noinspection PyAttributeOutsideInit
    @Node.once
    def make_renderable(self, ctx):
        self.prog = get_cylinder_program()
        self.prog['r1'] = self.r_base
        self.prog['r2'] = self.r_tip

        self.vbo_vertices = ctx.buffer(self.lines_current.astype('f4').tobytes())
        self.vbo_colors = ctx.buffer(self.colors.astype('f4').tobytes())
        self.vao.buffer(self.vbo_vertices, '3f', ['in_position'])
        self.vao.buffer(self.vbo_colors, '4f', ['in_color'])

    def render(self, camera, **kwargs):
        self.set_camera_matrices(self.prog, camera, **kwargs)
        set_lights_in_program(self.prog, kwargs['lights'], kwargs['shadows_enabled'])
        set_material_properties(self.prog, self.material)
        self.vao.render(self.prog)



"""
Copyright (C) 2022  ETH Zurich, Manuel Kaufmann, Velko Vechev

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import numpy as np
import joblib
import os

from aitviewer.configuration import CONFIG as C
from aitviewer.scene.camera_utils import look_at
from aitviewer.scene.camera_utils import orthographic_projection
from aitviewer.scene.camera_utils import perspective_projection
from trimesh.transformations import rotation_matrix

def _transform_vector(transform, vector):
    """Apply affine transformation (4-by-4 matrix) to a 3D vector."""
    return (transform @ np.concatenate([vector, np.array([1])]))[:3]


class PinholeCamera(object):
    """
    Your classic pinhole camera.
    """

    def __init__(self, fov=45, orthographic=None, znear=C.znear, zfar=C.zfar):
        self.fov = fov
        self.is_ortho = orthographic is not None
        self.ortho_size = 1.0 if orthographic is None else orthographic

        # Default camera settings.
        self.position = np.array([0.0, 0.0, 2.5])
        self.target = np.array([0.0, 0.0, 0.0])
        self.up = np.array([0.0, 1.0, 0.0])

        self.ZOOM_FACTOR = 4
        self.ROT_FACTOR = 0.0025
        self.PAN_FACTOR = 0.01

        self.near = znear
        self.far = zfar

    @property
    def dir(self):
        dir = self.target - self.position
        return dir / np.linalg.norm(dir)

    def save_cam(self):
        """Saves the current camera parameters"""
        cam_dir = C.export_dir + '/camera_params/'
        if not os.path.exists(cam_dir):
            os.makedirs(cam_dir)

        cam_dict = {}
        cam_dict['position'] = self.position
        cam_dict['target'] = self.target
        cam_dict['up'] = self.up
        cam_dict['ZOOM_FACTOR'] = self.ZOOM_FACTOR
        cam_dict['ROT_FACTOR'] = self.ROT_FACTOR
        cam_dict['PAN_FACTOR'] = self.PAN_FACTOR
        cam_dict['near'] = self.near
        cam_dict['far'] = self.far

        joblib.dump(cam_dict, cam_dir+'cam_params.pkl')

    def load_cam(self):
        """Loads the camera parameters"""
        cam_dir = C.export_dir + '/camera_params/'
        if not os.path.exists(cam_dir):
            print('camera config does not exist')
        else:
            cam_dict = joblib.load(cam_dir + 'cam_params.pkl')
            self.position = cam_dict['position']
            self.target = cam_dict['target']
            self.up = cam_dict['up']
            self.ZOOM_FACTOR = cam_dict['ZOOM_FACTOR']
            self.ROT_FACTOR = cam_dict['ROT_FACTOR']
            self.PAN_FACTOR = cam_dict['PAN_FACTOR']
            self.near = cam_dict['near']
            self.far = cam_dict['far']

    def update_matrices(self, width, height):
        #Compute projection matrix
        if self.is_ortho:
            yscale = self.ortho_size
            xscale = width / height * yscale
            P = orthographic_projection(xscale, yscale, self.near, self.far)
        else:
            P = perspective_projection(np.deg2rad(self.fov), width / height, self.near, self.far)
        
        #Compute view matrix
        V = look_at(self.position, self.target, self.up)

        #Update camera matrices
        self.projection_matrix = P.astype('f4')
        self.view_matrix = V.astype('f4')
        self.view_projection_matrix = np.matmul(P, V).astype('f4')

    def get_projection_matrix(self):
        """
        Returns the matrix that projects 3D coordinates in camera space to the image plane.
        """
        if self.projection_matrix is None:
            raise ValueError("update_matrices() must be called before to update the projection matrix")
        return self.projection_matrix
    
    def get_view_matrix(self):
        """
        Returns the matrix that projects 3D coordinates in camera space to the image plane.
        """
        if self.view_matrix is None:
            raise ValueError("update_matrices() must be called before to update the view matrix")
        return self.view_matrix


    def get_view_projection_matrix(self):
        """
        Returns the view-projection matrix, i.e. the 4x4 matrix that maps from homogenous world coordinates to image
        space.
        """
        if self.view_projection_matrix is None:
            raise ValueError("update_matrices() must be called before to update the view-projection matrix")
        return self.view_projection_matrix

    def dolly_zoom(self, speed):
        """Zoom by moving the camera along its view direction."""
        # We update both the orthographic and perspective projection so that the transition is seamless when
        # transitioning between them.
        self.ortho_size -= 0.1 * np.sign(speed)
        self.ortho_size = max(0.0001, self.ortho_size)

        # Scale the speed in proportion to the norm (i.e. camera moves slower closer to the target)
        norm = np.linalg.norm(self.position - self.target)
        fwd = self.dir / np.linalg.norm(self.dir)

        # Adjust speed according to config
        speed *= C.camera_zoom_speed

        self.position += fwd * speed * norm

    def pan(self, mouse_dx, mouse_dy):
        """Move the camera in the image plane."""
        sideways = np.cross(self.dir, self.up)
        up = np.cross(sideways, self.dir)

        speed_x = mouse_dx * self.PAN_FACTOR
        speed_y = mouse_dy * self.PAN_FACTOR

        self.position -= sideways * speed_x
        self.target -= sideways * speed_x

        self.position += up * speed_y
        self.target += up * speed_y

    def rotate_azimuth_elevation(self, mouse_dx, mouse_dy):
        """Rotate the camera left-right and up-down (roll is not allowed)."""
        cam_pose = np.linalg.inv(self.view_matrix)

        z_axis = cam_pose[:3, 2]
        dot = np.dot(z_axis, self.up)
        rot = np.eye(4)

        # Avoid singularity when z axis of camera is aligned with the up axis of the scene.
        if not (mouse_dy > 0 and dot > 0 and 1 - dot < 0.001) and not (mouse_dy < 0 and dot < 0 and 1 + dot < 0.001):
            # We are either hovering exactly below or above the scene's target but we want to move away or we are
            # not hitting the singularity anyway.
            x_axis = cam_pose[:3, 0]
            rot_x = rotation_matrix(self.ROT_FACTOR * -mouse_dy, x_axis, self.target)
            rot = rot_x @ rot

        y_axis = cam_pose[:3, 1]
        x_speed = self.ROT_FACTOR / 10 if 1 - np.abs(dot) < 0.01 else self.ROT_FACTOR
        rot = rotation_matrix(x_speed * -mouse_dx, y_axis, self.target) @ rot

        self.position = _transform_vector(rot, self.position)

    def rotate_azimuth(self, angle):
        """Rotate around camera's up-axis by given angle (in radians)."""
        if np.abs(angle) < 1e-8:
            return
        cam_pose = np.linalg.inv(self.view_matrix)
        y_axis = cam_pose[:3, 1]
        rot = rotation_matrix(angle, y_axis, self.target)
        self.position = _transform_vector(rot, self.position)

    def get_ray(self, x, y, width, height):
        """Construct a ray going through the middle of the given pixel."""
        w, h = width, height

        # Pixel in (-1, 1) range.
        screen_x = (2 * (x + 0.5) / w - 1)
        screen_y = (1 - 2 * (y + 0.5) / h)

        # Scale to actual image plane size.
        scale = self.ortho_size if self.is_ortho else np.tan(np.deg2rad(self.fov) / 2)
        screen_x *= scale * w / h
        screen_y *= scale

        pixel_2d = np.array([screen_x, screen_y, 0 if self.is_ortho else -1])
        cam2world = np.linalg.inv(self.view_matrix)
        pixel_3d = _transform_vector(cam2world, pixel_2d)
        if self.is_ortho:
            ray_origin = pixel_3d
            ray_dir = self.dir
        else:
            eye_origin = np.zeros(3)
            ray_origin = _transform_vector(cam2world, eye_origin)
            ray_dir = pixel_3d - ray_origin
        ray_dir = ray_dir / np.linalg.norm(ray_dir)

        return ray_origin, ray_dir

    def gui(self, imgui):
        _, self.is_ortho = imgui.checkbox('Orthographic Camera', self.is_ortho)
        _, self.fov = imgui.slider_float('Camera FOV##fov', self.fov, 0.1, 180.0, '%.1f')



"""
Copyright (C) 2022  ETH Zurich, Manuel Kaufmann, Velko Vechev

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import numpy as np
import os

from aitviewer.scene.node import Node
from aitviewer.renderables.spheres import Spheres
from aitviewer.renderables.arrows import Arrows


class RigidBodies(Node):
    """
    A sequence of N 3D positions and orientations in space.
    """
    def __init__(self,
                 rb_pos,
                 rb_ori,
                 radius=0.02,
                 length=0.2,
                 radius_cylinder=None,
                 color=(0.0, 1.0, 0.5, 1.0),
                 **kwargs):
        """
        Initializer.
        :param rb_pos: A np array of shape (F, N, 3) containing N rigid-body centers over F time steps.
        :param rb_ori: A np array of shape (F, N, 3, 3) containing N rigid-body orientations over F time steps.
        :param radius: Radius of the sphere at the origin of the rigid body.
        :param length: Length of arrows representing the orientation of the rigid body.
        :param radius_cylinder: Radius of the cylinder representing the orientation, default is length / 50
        :param color: Color of the rigid body centers (4-tuple).
        """
        self.rb_pos = rb_pos[np.newaxis] if rb_pos.ndim == 2 else rb_pos
        self.rb_ori = rb_ori[np.newaxis] if rb_ori.ndim == 3 else rb_ori
        super(RigidBodies, self).__init__(n_frames=self.rb_pos.shape[0], color=color, **kwargs)

        self.radius = radius
        self.length = length

        self.spheres = Spheres(rb_pos, radius=radius, color=color, position=self.position)
        self._add_node(self.spheres, has_gui=False, show_in_hierarchy=False)

        self.coords = []
        r_base = radius_cylinder or length / 50
        r_head = r_base * 2
        c = [0.0, 0.0, 0.0, 1.0]
        for i in range(3):
            line = self.rb_ori[..., :, i]
            line = line / np.linalg.norm(line, axis=-1, keepdims=True) * length
            color = c.copy()
            color[i] = 1.0
            axs = Arrows(self.rb_pos, self.rb_pos + line, r_base=r_base, r_head=r_head, color=tuple(color))
            axs.position = self.position
            self._add_node(axs, has_gui=False, show_in_hierarchy=False)
            self.coords.append(axs)

    @Node.color.setter
    def color(self, color):
        self.material.color = color
        self.spheres.color = color

    def redraw(self, **kwargs):
        self.spheres.sphere_positions = self.rb_pos
        self.spheres.redraw(**kwargs)

        for i in range(3):
            line = self.rb_ori[..., :, i]
            line = line / np.linalg.norm(line, axis=-1, keepdims=True) * self.length
            axs = self.coords[i]
            axs.position = self.position
            axs.origins = self.rb_pos
            axs.tips = self.rb_pos + line
            axs.redraw(**kwargs)

    def gui(self, imgui):
        super(RigidBodies, self).gui(imgui)
        # Scale controls
        u, scale = imgui.drag_float('Sphere Radius##radius{}'.format(self.unique_name), self.spheres.radius,
                                    0.01, min_value=0.001, max_value=10.0, format='%.3f')
        if u:
            self.spheres.radius = scale
            self.redraw()



"""
Copyright (C) 2022  ETH Zurich, Manuel Kaufmann, Velko Vechev

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import cv2

from aitviewer.streamables.streamable import Streamable
from aitviewer.scene.node import Node
from aitviewer.shaders import get_screen_texture_program
from moderngl_window import geometry


class Webcam(Streamable):
    """
    Renders webcam stream to a quad. Quad is positioned in screen coordinates.
    """

    def __init__(self, src=0, size=(2.0, 2.0), pos=(0.0, 0.0), transparency=1.0, icon="\u0088", **kwargs):
        '''
        :param src: integer denotes device source id, i.e. webcam 0.
        '''

        super(Webcam, self).__init__(icon=icon, **kwargs)

        # Capture source
        self._cap = None

        # Set after cv reads video file or webcam
        self.width = None
        self.height = None
        self.pos = pos
        self.size = size
        self.fps = None
        self.frame_count = None
        self.transparency = transparency
        self.src = src

        # Render into a quad in screen space (z=0)
        self._texture = None

    @Node.once
    def make_renderable(self, ctx):
        self.ctx = ctx
        self.quad = geometry.quad_2d(pos=self.pos, size=self.size, normals=False)  # (2,2) is Full Screen i.e. -1 to 1 in x/y
        self.prog = get_screen_texture_program()

    def capture(self):
        ret, frame = self._cap.read()
        if ret:
            frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
            self._texture.write(frame)

    def render(self, camera, **kwargs):
        self._texture.use(0)
        self.prog['transparency'].value = self.transparency
        self.quad.render(self.prog)

    @property
    def enabled(self):
        return self._cap is not None and self._cap.isOpened()

    @enabled.setter
    def enabled(self, enabled):
        # Not running - start
        if enabled and not self.enabled:
            self.start()
        # Running - stop
        elif not enabled and self.enabled:
            self.stop()

    def start(self):
        # Capture from webcam / device source
        self._cap = cv2.VideoCapture(self.src)
        self.width = int(self._cap.get(cv2.CAP_PROP_FRAME_WIDTH))
        self.height = int(self._cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
        self.fps = self._cap.get(cv2.CAP_PROP_FPS)

        # Set W/H Manually
        # self._cap.set(cv2.CAP_PROP_FRAME_WIDTH, 1280)
        # self._cap.set(cv2.CAP_PROP_FRAME_HEIGHT, 720)

        # Initialize texture to
        self._texture = self.ctx.texture((self.width, self.height), 3)

        # Check if the webcam is opened correctly
        if not self._cap.isOpened():
            raise IOError("Cannot open source")

    def stop(self):
        self._cap.release()
        self._texture = None

    def gui(self, imgui):
        if self.enabled:
            imgui.text("Clip Info: {}x{} @ {:.2f} fps".format(self.width, self.height, self.fps))
        _, self.transparency = imgui.slider_float('Opacity##opacity_'.format(self.name), self.transparency, 0.0, 1.0, '%.2f')


"""
Copyright (C) 2022  ETH Zurich, Manuel Kaufmann, Velko Vechev

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
from functools import lru_cache
import numpy as np
from aitviewer.renderables.lines import Lines

from aitviewer.scene.node import Node
from aitviewer.scene.camera_utils import look_at
from aitviewer.scene.camera_utils import orthographic_projection


class Light(Node):
    """Simple point light."""

    def __init__(self, intensity_diffuse=1.0, intensity_ambient=1.0, shadow_enabled=True, **kwargs):
        super(Light, self).__init__(**kwargs)

        self.intensity_ambient = intensity_ambient
        self.intensity_diffuse = intensity_diffuse

        self.shadow_enabled = shadow_enabled
        self.shadow_map = None

        self.shadow_map_size = 15.0
        self.shadow_map_near = 5.0
        self.shadow_map_far = 50.0
        
        self._debug_lines = None
        self._show_debug_lines = False

    def create_shadowmap(self, ctx):
        if self.shadow_map is None:
            shadow_map_size = 8192, 8192
            
            # Setup shadow mapping
            self.shadow_map = ctx.depth_texture(shadow_map_size)
            self.shadow_map.compare_func = '>'
            self.shadow_map.repeat_x = False
            self.shadow_map.repeat_y = False
            self.shadow_map_framebuffer = ctx.framebuffer(depth_attachment=self.shadow_map)
    
    def use(self, ctx):
        if not self.shadow_map:
            self.create_shadowmap(ctx)

        self.shadow_map_framebuffer.clear()
        self.shadow_map_framebuffer.use()

    @staticmethod
    @lru_cache()
    def _compute_light_matrix(position, size, near, far):
        P = orthographic_projection(size, size, near, far)
        V = look_at(np.array(position), np.array([0.0, 0.0, 0.0]), np.array([0.0, 1.0, 0.0]))
        return (P @ V).astype('f4')

    def mvp(self):
        """Return a model-view-projection matrix to project vertices into the view of the light."""
        return self._compute_light_matrix(tuple(self.position), self.shadow_map_size, self.shadow_map_near, self.shadow_map_far)
    
    def _update_debug_lines(self):
        lines = np.array([
            [-1, -1, -1], [-1,  1, -1],
            [-1, -1,  1], [-1,  1,  1],
            [ 1, -1, -1], [ 1,  1, -1],
            [ 1, -1,  1], [ 1,  1,  1],
            
            [-1, -1, -1], [-1, -1, 1],
            [-1,  1, -1], [-1,  1, 1],
            [ 1, -1, -1], [ 1, -1, 1],
            [ 1,  1, -1], [ 1,  1, 1],
            
            [-1, -1, -1], [ 1, -1, -1],
            [-1, -1,  1], [ 1, -1,  1],
            [-1,  1, -1], [ 1,  1, -1],
            [-1,  1,  1], [ 1,  1,  1],
        ])

        world_from_ndc = np.linalg.inv(self.mvp())
        lines = np.apply_along_axis(lambda x: (world_from_ndc @ np.append(x, 1.0))[:3] - self.position, 1, lines)

        if self._debug_lines is None:
            self._debug_lines = Lines(lines, position=self.position, r_base=0.2, mode='lines', cast_shadow=False)
            self.add(self._debug_lines)
        else:
            self._debug_lines.lines = lines
            self._debug_lines.redraw()

    @Node.position.setter
    def position(self, position):
        self._position = position
        for n in self.nodes:
            n.position = position
        self._update_debug_lines()

    def redraw(self, **kwargs):
        if self._debug_lines:
            self._debug_lines.redraw(**kwargs)

    def gui(self, imgui):
        self.gui_position(imgui)
        self.gui_material(imgui, show_advanced=False)

        # Light controls
        _, self.intensity_ambient = imgui.drag_float('Ambient##ambient', self.intensity_ambient, 0.01, min_value=0.0, max_value=1.0,
                                           format='%.2f')
        _, self.intensity_diffuse = imgui.drag_float('Diffuse##diffuse', self.intensity_diffuse, 0.01, min_value=0.0, max_value=1.0,
                                           format='%.2f')
                          
        _, self.shadow_enabled = imgui.checkbox('Enable Shadows', self.shadow_enabled)

        u_size, self.shadow_map_size = imgui.drag_float('Shadow Map Size', self.shadow_map_size, 0.1, format='%.2f', min_value=0.01, max_value=100.0)
        u_near, self.shadow_map_near = imgui.drag_float('Shadow Map Near', self.shadow_map_near, 0.1, format='%.2f', min_value=0.01, max_value=100.0)
        u_far, self.shadow_map_far  = imgui.drag_float('Shadow Map Far', self.shadow_map_far, 0.1, format='%.2f', min_value=0.01, max_value=100.0)

        u_show, self._show_debug_lines = imgui.checkbox('Show Shadow Map Frustum', self._show_debug_lines)

        if self._show_debug_lines:
            if self._debug_lines:
                self._debug_lines.enabled = True

            if u_size or u_near or u_far or u_show:
                self._update_debug_lines()
        else:
            if self._debug_lines:
                self._debug_lines.enabled = False
        
        


"""
Copyright (C) 2022  ETH Zurich, Manuel Kaufmann, Velko Vechev

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import numpy as np
import roma
import torch

from scipy.spatial.transform import Rotation as R
from scipy.spatial.transform import RotationSpline


def rot2aa_torch(rotation_matrices):
    """
    Convert rotation matrices to rotation vectors (angle-axis representation).
    :param rotation_matrices: A torch tensor of shape (..., 3, 3).
    :return: A torch tensor of shape (..., 3).
    """
    assert isinstance(rotation_matrices, torch.Tensor)
    return roma.rotmat_to_rotvec(rotation_matrices)


def aa2rot_torch(rotation_vectors):
    """
    Convert rotation vectors (angle-axis representation) to rotation matrices.
    :param rotation_vectors: A torch tensor of shape (..., 3).
    :return: A torch tensor of shape (..., 3, 3).
    """
    assert isinstance(rotation_vectors, torch.Tensor)
    return roma.rotvec_to_rotmat(rotation_vectors)


def rot2aa_numpy(rotation_matrices):
    """
    Convert rotation matrices to rotation vectors (angle-axis representation).
    :param rotation_matrices: A numpy array of shape (..., 3, 3).
    :return: A numpy array of shape (..., 3).
    """
    assert isinstance(rotation_matrices, np.ndarray)
    ori_shape = rotation_matrices.shape[:-2]
    rots = np.reshape(rotation_matrices, (-1, 3, 3))
    aas = R.as_rotvec(R.from_matrix(rots))
    rotation_vectors = np.reshape(aas, ori_shape + (3, ))
    return rotation_vectors


def aa2rot_numpy(rotation_vectors):
    """
    Convert rotation matrices to rotation vectors (angle-axis representation).
    :param rotation_vectors: A numpy array of shape (..., 3).
    :return: A numpy array of shape (..., 3, 3).
    """
    assert isinstance(rotation_vectors, np.ndarray)
    ori_shape = rotation_vectors.shape[:-1]
    aas = np.reshape(rotation_vectors, (-1, 3))
    rots = R.as_matrix(R.from_rotvec(aas))
    rotation_matrices = np.reshape(rots, ori_shape + (3, 3))
    return rotation_matrices


def interpolate_rotations(rotations, ts_in, ts_out):
    """
    Interpolate rotations given at timestamps `ts_in` to timestamps given at `ts_out`. This performs the equivalent
    of cubic interpolation in SO(3).
    :param rotations: A numpy array of rotations of shape (F, N, 3), i.e. rotation vectors.
    :param ts_in: Timestamps corresponding to the given rotations, len(ts_in) == F
    :param ts_out: The desired output timestamps.
    :return: A numpy array of shape (len(ts_out), N, 3).
    """
    out = []
    for j in range(rotations.shape[1]):
        rs = R.from_rotvec(rotations[:, j])
        spline = RotationSpline(ts_in, rs)
        rs_interp = spline(ts_out).as_rotvec()
        out.append(rs_interp[:, np.newaxis])
    return np.concatenate(out, axis=1)


def resample_rotations(rotations, fps_in, fps_out):
    """
    Resample a motion sequence from `fps_in` to `fps_out`.
    :param rotations: A numpy array of shape (F, N, 3), i.e. in angle-axis form.
    :param fps_in: The frequency of the input sequence.
    :param fps_out: The desired frequency of the output sequence.
    :return: A numpy array of shape (F', N, 3) where F is adjusted according to the new fps.
    """
    n_frames = rotations.shape[0]
    assert n_frames > 1, "We need at least two rotations for a resampling to make sense."
    duration = n_frames / fps_in
    ts_in = np.arange(0, duration, 1 / fps_in)[:n_frames]
    ts_out = np.arange(0, duration, 1 / fps_out)
    return interpolate_rotations(rotations, ts_in, ts_out)



"""
Copyright (C) 2022  ETH Zurich, Manuel Kaufmann, Velko Vechev

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import numpy as np

from aitviewer.scene.node import Node
from aitviewer.renderables.lines import Lines
from aitviewer.renderables.spheres import Spheres


class BoundingBoxes(Node):
    """
    Draw bounding boxes.
    """

    def __init__(self,
                 vertices,
                 thickness=0.005,
                 color=(0.0, 0.0, 1.0, 1.0),
                 **kwargs):
        """
        Initializer.
        :param vertices: Set of 3D coordinates as a np array of shape (N, 8, 3). The vertices will be connected in the
          following way: 0-1-2-3-0 (bottom) 4-5-6-7-4 (top) 0-4 1-5 2-6 3-7 (vertical connections between bottom
          and top).
        :param thickness: Line thickness.
        :param color: Color of the lines.
        """
        if not isinstance(vertices, np.ndarray):
            vertices = np.array(vertices)
        if len(vertices.shape) == 2:
            vertices = vertices[np.newaxis]
        else:
            assert len(vertices.shape) == 3
        assert vertices.shape[1] == 8
        super(BoundingBoxes, self).__init__(n_frames=len(vertices), color=color, **kwargs)

        self.vertices = vertices

        self.lines = Lines(lines=self._get_line_coords(), mode='lines', r_base=thickness, color=self.color,
                           position=self.position)
        self.spheres = Spheres(positions=self.vertices, radius=thickness, color=self.color)
        self._add_nodes(self.lines, self.spheres, show_in_hierarchy=False)

    @staticmethod
    def from_min_max_diagonal(v_min, v_max, **kwargs):
        """
        Create an axis-aligned bounding box from the 3D diagonal.
        :param v_min: np array of shape (N, 3).
        :param v_max: np array of shape (N, 3).
        :return: BoundingBoxes corresponding to the given diagonals.
        """
        vertices = np.zeros((v_min.shape[0], 8, 3), dtype=v_min.dtype)
        vertices[:, 0:4] = v_min[:, np.newaxis]
        vertices[:, 1, 0] = v_max[:, 0]
        vertices[:, 2, 0:2] = v_max[:, 0:2]
        vertices[:, 3, 1] = v_max[:, 1]

        vertices[:, 4:] = v_max[:, np.newaxis]
        vertices[:, 4, 0:2] = v_min[:, 0:2]
        vertices[:, 7, 0] = v_min[:, 0]
        vertices[:, 5, 1] = v_min[:, 1]

        return BoundingBoxes(vertices, **kwargs)

    def _get_line_coords(self):
        lines = np.zeros((self.n_frames, 12*2, 3), dtype=self.vertices.dtype)

        # Bottom 0-1-2-3-0.
        lines[:, 0:2] = self.vertices[:, 0:2]
        lines[:, 2:4] = self.vertices[:, 1:3]
        lines[:, 4:6] = self.vertices[:, 2:4]
        lines[:, 6:8] = self.vertices[:, [3, 0]]

        # Top 4-5-6-7-4.
        lines[:, 8:10] = self.vertices[:, 4:6]
        lines[:, 10:12] = self.vertices[:, 5:7]
        lines[:, 12:14] = self.vertices[:, 6:8]
        lines[:, 14:16] = self.vertices[:, [7, 4]]

        # Vertical Connections.
        lines[:, 16:18] = self.vertices[:, [0, 4]]
        lines[:, 18:20] = self.vertices[:, [1, 5]]
        lines[:, 20:22] = self.vertices[:, [2, 6]]
        lines[:, 22:24] = self.vertices[:, [3, 7]]

        return lines

    @Node.color.setter
    def color(self, color):
        self.material.color = color
        self.lines.color = color
        self.spheres.color = color



"""
Copyright (C) 2022  ETH Zurich, Manuel Kaufmann, Velko Vechev

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import numpy as np

from aitviewer.configuration import CONFIG as C
from moderngl_window.opengl.vao import VAO


class Material(object):
    """Per object material properties."""

    def __init__(self,
                 diffuse=0.5,
                 ambient=0.5,
                 specular=0.5,
                 color=(0.5, 0.5, 0.5, 1.0),
                 show_edges=True,
                 ):
        """
        :param diffuse: diffuse coefficient in Phong shading model
        :param ambient: ambient coefficient in Phong shading model
        :param specular: specular coefficient in Phong shading model
        :param color: (R,G,B,A) 0-1 formatted color value
        :param show_edges: Whether to show edges in edge view
        """
        assert len(color) == 4

        self.diffuse = diffuse
        self.ambient = ambient
        self.specular = specular
        self._show_edges = show_edges
        self._color = color

    @property
    def color(self):
        return self._color

    @color.setter
    def color(self, color):
        self._color = color



"""
Copyright (C) 2022  ETH Zurich, Manuel Kaufmann, Velko Vechev

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
from abc import ABC
import collections
import numpy as np
import smplx
import torch
import torch.nn as nn
import trimesh

from aitviewer.utils.utils import compute_vertex_and_face_normals_torch
from aitviewer.configuration import CONFIG as C
from aitviewer.utils.so3 import aa2rot_torch as aa2rot
from aitviewer.utils.so3 import rot2aa_torch as rot2aa


class SMPLLayer(nn.Module, ABC):
    """A wrapper for the various SMPL body models."""

    def __init__(self, model_type='smpl', gender='neutral', num_betas=10, device=C.device, dtype=C.f_precision,
                 **smpl_model_params):
        """
        Initializer.
        :param model_type: Which type of SMPL model to load, currently SMPL, SMPL-H and SMPL-X are supported.
        :param gender: Which gender to load.
        :param num_betas: Number of shape components.
        :param device: CPU or GPU.
        :param dtype: The pytorch floating point data type.
        :param smpl_model_params: Other keyword arguments that can be passed to smplx.create.
        """
        assert model_type in ['smpl', 'smplh', 'smplx', 'mano', 'flame']
        assert gender in ['male', 'female', 'neutral']
        if model_type == 'smplh' and gender == 'neutral':
            gender = 'female'  # SMPL-H has no neutral gender.

        super(SMPLLayer, self).__init__()
        self.num_betas = num_betas

        smpl_model_params['use_pca'] = smpl_model_params.get('use_pca', False)
        smpl_model_params['flat_hand_mean'] = smpl_model_params.get('flat_hand_mean', True)

        self.bm = smplx.create(C.smplx_models, model_type=model_type,
                               num_betas=self.num_betas, gender=gender, **smpl_model_params)
        self.bm.to(device=device, dtype=dtype)

        self._parents = None
        self._children = None
        self._closest_joints = None
        self._vertex_faces = None
        self._faces = None

    @property
    def faces(self):
        """Return the definition of the faces."""
        if self._faces is None:
            self._faces = self.bm.f.to(dtype=torch.int32)
        return self._faces

    @property
    def parents(self):
        """Return how the joints are connected in the kinematic chain where parents[i, 0] is the parent of
        joint parents[i, 1]."""
        if self._parents is None:
            self._parents = self.bm.kintree_table.transpose(0, 1).cpu().numpy()
        return self._parents

    @property
    def joint_children(self):
        """Return the children of each joint in the kinematic chain."""
        if self._children is None:
            self._children = collections.defaultdict(list)
            for bone in self.parents:
                if bone[0] != -1:
                    self._children[bone[0]].append(bone[1])
        return self._children

    def vertex_faces(self, vertices):
        """Return a matrix that returns a list of faces each vertex is contributing to. `vertices` should have
        have shape (V, 3)."""
        if self._vertex_faces is None:
            mesh = trimesh.Trimesh(vertices.detach().cpu().numpy(), self.faces.cpu().numpy(), process=False)
            self._vertex_faces = torch.from_numpy(np.copy(mesh.vertex_faces)).to(dtype=torch.long, device=vertices.device)
        return self._vertex_faces

    def vertex_normals(self, vertices, output_vertex_ids=None):
        """
        Return the unnormalized vertex normals at the provided vertex IDs.
        :param vertices: A tensor of shape (N, V, 3).
        :param output_vertex_ids: An optional list of integers indexing into the 2nd dimension of `vertices`.
        :return: A tensor of shape (N, V', 3) where V' is either V or len(output_vertex_ids).
        """
        normals, _ = compute_vertex_and_face_normals_torch(vertices,
                                                           self.faces,
                                                           self.vertex_faces(vertices[0]))
        if output_vertex_ids is not None:
            return normals[:, output_vertex_ids]
        else:
            return normals

    def skeletons(self):
        """Return how the joints are connected in the kinematic chain where skeleton[0, i] is the parent of
        joint skeleton[1, i]."""
        parents = torch.stack([self.bm.parents, torch.arange(0, len(self.bm.parents), device=self.bm.parents.device)])
        return {'all': parents, 'body': parents[:, :self.bm.NUM_BODY_JOINTS + 1],
                'hands': parents[:, self.bm.NUM_BODY_JOINTS + 1:]}

    def fk(self, poses_body, betas, poses_root=None, trans=None, normalize_root=False,
           poses_left_hand=None, poses_right_hand=None,
           poses_jaw=None, poses_leye=None, poses_reye=None, expression=None):
        """
        Convert body pose data (joint angles and shape parameters) to positional data (joint and mesh vertex positions).
        :param poses_body: A tensor of shape (N, N_JOINTS*3), i.e. joint angles in angle-axis format. This contains all
          body joints which are not the root, i.e. possibly including hands and face depending on the underlying body
          model.
        :param betas: A tensor of shape (N, N_BETAS) containing the betas/shape parameters, i.e. shape parameters can
          differ for every sample. If N_BETAS > self.num_betas, the excessive shape parameters will be ignored.
        :param poses_root: Orientation of the root or None. If specified expected shape is (N, 3).
        :param trans: Translation of the root or None. If specified expected shape is (N, 3).
        :param normalize_root: If set, it will normalize the root such that its orientation is the identity in the
          first frame and its position starts at the origin.
        :param poses_left_hand: A tensor of shape (N, N_JOINTS_HANDS*3) or None. Only relevant if this body model
          supports hands.
        :param poses_right_hand: A tensor of shape (N, N_JOINTS_HANDS*3) or None. Only relevant if this body model
          supports hands.
        :return: The resulting vertices and joints.
        """
        assert poses_body.shape[1] == self.bm.NUM_BODY_JOINTS*3

        has_hands = hasattr(self.bm, 'NUM_HAND_JOINTS')
        has_face = hasattr(self.bm, 'NUM_FACE_JOINTS')
        if has_hands:
            if self.bm.use_pca:
                assert poses_left_hand is None or poses_left_hand.shape[1] == 6
                assert poses_right_hand is None or poses_right_hand.shape[1] == 6
            else:
                assert poses_left_hand is None or poses_left_hand.shape[1] == self.bm.NUM_HAND_JOINTS*3
                assert poses_right_hand is None or poses_right_hand.shape[1] == self.bm.NUM_HAND_JOINTS*3

        batch_size = poses_body.shape[0]
        device = poses_body.device

        if poses_root is None:
            poses_root = torch.zeros([batch_size, 3]).to(dtype=poses_body.dtype, device=device)
        if trans is None:
            trans = torch.zeros([batch_size, 3]).to(dtype=poses_body.dtype, device=device)

        if has_hands and poses_left_hand is None:
            poses_left_hand = torch.zeros([batch_size, self.bm.NUM_HAND_JOINTS*3]).to(dtype=poses_body.dtype,
                                                                                      device=device)
        if has_hands and poses_right_hand is None:
            poses_right_hand = torch.zeros([batch_size, self.bm.NUM_HAND_JOINTS * 3]).to(dtype=poses_body.dtype,
                                                                                         device=device)
        if has_face and poses_jaw is None:
            poses_jaw = torch.zeros([batch_size, 3]).to(dtype=poses_body.dtype, device=device)
        if has_face and poses_leye is None:
            poses_leye = torch.zeros([batch_size, 3]).to(dtype=poses_body.dtype, device=device)
        if has_face and poses_reye is None:
            poses_reye = torch.zeros([batch_size, 3]).to(dtype=poses_body.dtype, device=device)
        if has_face and expression is None:
            expression = torch.zeros([batch_size, self.bm.num_expression_coeffs]).to(dtype=poses_body.dtype,
                                                                                     device=device)

        # Batch shapes if they don't match batch dimension.
        if len(betas.shape) == 1 or betas.shape[0] == 1:
            betas = betas.repeat(poses_body.shape[0], 1)
        betas = betas[:, :self.num_betas]

        if normalize_root:
            # Make everything relative to the first root orientation.
            root_ori = aa2rot(poses_root)
            first_root_ori = torch.inverse(root_ori[0:1])
            root_ori = torch.matmul(first_root_ori, root_ori)
            poses_root = rot2aa(root_ori)
            trans = torch.matmul(first_root_ori.unsqueeze(0), trans.unsqueeze(-1)).squeeze()
            trans = trans - trans[0:1]

        bm_output = self.bm(body_pose=poses_body, betas=betas, global_orient=poses_root, transl=trans,
                            left_hand_pose=poses_left_hand, right_hand_pose=poses_right_hand,
                            jaw_pose=poses_jaw, leye_pose=poses_leye, reye_pose=poses_reye, expression=expression)
        return bm_output['vertices'], bm_output['joints']

    def forward(self, *args, **kwargs):
        """
        Forward pass using forward kinematics
        """
        return self.fk(*args, **kwargs)



from moderngl_window.timers.base import BaseTimer
import time
from typing import Tuple

"""
Timer class adapted from moderngl_window.timers.clock.Timer.
Uses time.perf_counter() instead of time.time() for higher resolution timestamps.
"""
class PerfTimer(BaseTimer):
    """Timer based on python ``time``."""

    def __init__(self, **kwargs):
        self._start_time = None
        self._stop_time = None
        self._pause_time = None
        self._last_frame = None
        self._offset = 0

    @property
    def is_paused(self) -> bool:
        """bool: The pause state of the timer"""
        return self._pause_time is not None

    @property
    def is_running(self) -> bool:
        """bool: Is the timer currently running?"""
        return self._pause_time is None

    @property
    def time(self) -> float:
        """Get or set the current time.
        This can be used to jump around in the timeline.

        Returns:
            The current time in seconds
        """
        if self._start_time is None:
            return 0.0

        if self.is_paused:
            return self._pause_time - self._offset - self._start_time

        return time.perf_counter() - self._start_time - self._offset

    @time.setter
    def time(self, value: float):
        if value < 0:
            value = 0

        self._offset += self.time - value

    def next_frame(self) -> Tuple[float, float]:
        """
        Get the time and frametime for the next frame.
        This should only be called once per frame.

        Returns:
            Tuple[float, float]: current time and frametime
        """
        current = self.time
        delta, self._last_frame = current - self._last_frame, current
        return current, delta

    def start(self):
        """Start the timer by recoding the current ``time.perf_counter()``
        preparing to report the number of seconds since this timestamp.
        """
        if self._start_time is None:
            self._start_time = time.perf_counter()
            self._last_frame = 0.0
        else:
            self._offset += time.perf_counter() - self._pause_time
            self._pause_time = None

    def pause(self):
        """Pause the timer by setting the internal pause time using ``time.perf_counter()``"""
        self._pause_time = time.perf_counter()

    def toggle_pause(self):
        """Toggle the paused state"""
        if self.is_paused:
            self.start()
        else:
            self.pause()

    def stop(self) -> Tuple[float, float]:
        """
        Stop the timer. Should only be called once when stopping the timer.

        Returns:
            Tuple[float, float]: Current position in the timer, actual running duration
        """
        self._stop_time = time.perf_counter()
        return (
            self._stop_time - self._start_time - self._offset,
            self._stop_time - self._start_time,
        )



"""
Copyright (C) 2022  ETH Zurich, Manuel Kaufmann, Velko Vechev

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import moderngl
import numpy as np

from aitviewer.scene.material import Material
from aitviewer.scene.node import Node
from aitviewer.shaders import get_smooth_lit_with_edges_program
from aitviewer.utils import set_lights_in_program
from aitviewer.utils import set_material_properties


def _create_spheres(radius=1.0, rings=16, sectors=32, n_spheres=1, create_faces=True):
    """
    Create a sphere centered at the origin. This is a port of moderngl-window's geometry.sphere() function, but it
    returns the vertices, normals, and faces explicitly instead of directly storing them in a VAO.
    :param radius: Radius of the sphere.
    :param rings: Longitudinal resolution.
    :param sectors: Latitudinal resolution.
    :param n_spheres: How many spheres to create.
    :param create_faces: Whether or not to create and return faces.
    :return: vertices, normals, and faces of the sphere.
    """
    R = 1.0 / (rings - 1)
    S = 1.0 / (sectors - 1)

    vertices = np.zeros([n_spheres, rings * sectors, 3])
    normals = np.zeros([n_spheres, rings * sectors, 3])

    v, n = 0, 0
    for r in range(rings):
        for s in range(sectors):
            y = np.sin(-np.pi / 2 + np.pi * r * R)
            x = np.cos(2 * np.pi * s * S) * np.sin(np.pi * r * R)
            z = np.sin(2 * np.pi * s * S) * np.sin(np.pi * r * R)

            vertices[:, v] = np.array([x, y, z]) * radius
            normals[:, n] = np.array([x, y, z])

            v += 1
            n += 1

    if create_faces:
        faces = np.zeros([n_spheres, rings * sectors * 2, 3], dtype=np.int32)
        i = 0
        for r in range(rings - 1):
            for s in range(sectors - 1):
                faces[:, i] = np.array([r * sectors + s,
                                        (r + 1) * sectors + (s + 1),
                                        r * sectors + (s + 1)])
                faces[:, i + 1] = np.array([r * sectors + s,
                                            (r + 1) * sectors + s,
                                            (r + 1) * sectors + (s + 1)])
                i += 2
    else:
        faces = None

    return {'vertices': vertices, 'normals': normals, 'faces': faces}


class Spheres(Node):
    """Render some simple spheres."""

    def __init__(self,
                 positions,
                 radius=0.01,
                 color=(0.0, 0.0, 1.0, 1.0),
                 rings=16,
                 sectors=32,
                 **kwargs):
        """
        Initializer.
        :param positions: A numpy array of shape (F, N, 3) or (N, 3) containing N sphere positions for F time steps.
        :param radius: Radius of the spheres.
        :param color: Color of the spheres.
        :param rings: Longitudinal resolution.
        :param sectors: Latitudinal resolution.
        :param kwargs: Remaining parameters.
        """
        if len(positions.shape) == 2:
            positions = positions[np.newaxis]
        assert len(positions.shape) == 3

        # Define a default material in case there is None.
        kwargs['material'] = kwargs.get('material', Material(color=color, ambient=0.2))
        super(Spheres, self).__init__(n_frames=positions.shape[0], **kwargs)

        self.sphere_positions = positions
        self.n_spheres = positions.shape[1]
        self.spheres_data = _create_spheres(radius=1.0, rings=rings, sectors=sectors, n_spheres=self.n_spheres)

        self.n_vertices = self.spheres_data['vertices'].shape[1]
        self.sphere_vertices = np.reshape(self.spheres_data['vertices'], [-1, 3])
        self.sphere_normals = np.reshape(self.spheres_data['normals'], [-1, 3])
        sphere_faces = [self.spheres_data['faces'][0]] + [self.spheres_data['faces'][i] + i * self.n_vertices for i in
                                                          range(1, self.n_spheres)]
        self.sphere_faces = np.concatenate(sphere_faces)
        self.radius = radius

    @property
    def vertex_colors(self):
        return np.full((self.n_spheres * self.n_vertices, 4), self.color)

    # noinspection PyAttributeOutsideInit
    @Node.once
    def make_renderable(self, ctx):
        self.prog = get_smooth_lit_with_edges_program()
        self.vbo_vertices = ctx.buffer(self.sphere_vertices.astype('f4').tobytes())
        self.vbo_colors = ctx.buffer(self.vertex_colors.astype('f4').tobytes())
        self.vbo_normals = ctx.buffer(self.sphere_normals.squeeze().astype('f4').tobytes())
        self.vbo_indices = ctx.buffer(self.sphere_faces.tobytes())

        self.mesh_vao = ctx.vertex_array(self.prog,
                                         [(self.vbo_vertices, '3f4 /v', 'in_position'),
                                          (self.vbo_colors, '4f4 /v', 'in_color'),
                                          (self.vbo_normals, '3f4 /v', 'in_normal')],
                                         self.vbo_indices)
        self.is_renderable = True
        self.redraw()

    def on_frame_update(self):
        self.redraw()

    def redraw(self, **kwargs):
        """Upload the current frame data to the GPU for rendering."""
        if not self.is_renderable:
            return

        current_pos = self.sphere_positions[self.current_frame_id]
        vertices = np.reshape(self.sphere_vertices, [-1, self.n_vertices, 3]) * self.radius + current_pos[:, np.newaxis]
        self.vbo_vertices.write(np.reshape(vertices, [-1, 3]).astype('f4').tobytes())
        self.vbo_colors.write(self.vertex_colors.astype('f4').tobytes())

    def render(self, camera, **kwargs):
        self.set_camera_matrices(self.prog, camera, **kwargs)
        set_lights_in_program(self.prog, kwargs['lights'], kwargs['shadows_enabled'])
        set_material_properties(self.prog, self.material)
        self.prog['draw_edges'].value = 1.0 if kwargs['draw_edges'] and self.material._show_edges else 0.0
        self.prog['norm_coloring'].value = False
        self.prog['win_size'].value = kwargs['window_size']
        self.mesh_vao.render(moderngl.TRIANGLES)

    @Node.color.setter
    def color(self, color):
        self.material.color = color
        self.redraw()

    def gui_scale(self, imgui):
        # Scale controls
        u, scale = imgui.drag_float('Radius##radius{}'.format(self.unique_name), self.radius, 0.01,
                                    min_value=0.001,
                                    max_value=10.0, format='%.3f')
        if u:
            self.radius = scale
            self.redraw()



"""
Copyright (C) 2022  ETH Zurich, Manuel Kaufmann, Velko Vechev

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import os

import moderngl
from moderngl_window.opengl.vao import VAO
import numpy as np
import trimesh
import tqdm
import re
import pickle

from functools import lru_cache
from PIL import Image

from aitviewer.scene.node import Node
from aitviewer.shaders import get_smooth_lit_with_edges_program
from aitviewer.shaders import get_flat_lit_with_edges_program
from aitviewer.shaders import get_smooth_lit_texturized_program
from aitviewer.utils.decorators import hooked
from aitviewer.utils.utils import compute_vertex_and_face_normals
from aitviewer.utils import set_lights_in_program
from aitviewer.utils import set_material_properties
from trimesh.triangles import points_to_barycentric


class Meshes(Node):
    """A sequence of triangle meshes. This assumes that the mesh topology is fixed over the sequence."""

    def __init__(self,
                 vertices,
                 faces,
                 vertex_normals=None,
                 face_normals=None,
                 vertex_colors=None,
                 face_colors=None,
                 uv_coords=None,
                 path_to_texture=None,
                 texture_alpha=1.0,
                 cast_shadow=True,
                 **kwargs):
        """
        Initializer.
        :param vertices: A np array of shape (N, V, 3) or (V, 3).
        :param faces: A np array of shape (F, 3).
        :param vertex_normals: A np array of shape (N, V, 3). If not provided, the vertex normals will be computed,
          which incurs some overhead.
        :param face_normals: A np array of shape (N, F, 3). If not provided, the face normals will be computed, which
          incurs some overhead.
        :param vertex_colors: A np array of shape (N, V, 4) overriding the uniform color.
        :param face_colors: A np array of shape (N, F, 4) overriding the uniform or vertex colors.
        :param uv_coords: A np array of shape (V, 2) if the mesh is to be textured.
        :param path_to_texture: Path to an image file that serves as the texture.
        :param texture_alpha: Set transparency for texture.
        """
        if len(vertices.shape) == 2 and vertices.shape[-1] == 3:
            vertices = vertices[np.newaxis]
        assert len(vertices.shape) == 3
        assert len(faces.shape) == 2
        super(Meshes, self).__init__(n_frames=vertices.shape[0], **kwargs)

        self._vertices = vertices
        self.faces = faces.astype(np.int32)

        def _maybe_unsqueeze(x):
            return x[np.newaxis] if x is not None and x.ndim == 2 else x

        self._vertex_normals = _maybe_unsqueeze(vertex_normals)
        self._face_normals = _maybe_unsqueeze(face_normals)
        self.vertex_colors = _maybe_unsqueeze(vertex_colors)
        self.face_colors = _maybe_unsqueeze(face_colors)

        # Texture handling.
        self.has_texture = uv_coords is not None
        self.uv_coords = uv_coords

        if self.has_texture:
            self.use_pickle_texture = path_to_texture.endswith((".pickle", "pkl"))
            if self.use_pickle_texture:
                self.texture_image = pickle.load(open(path_to_texture, "rb"))
            else:
                self.texture_image = Image.open(path_to_texture).transpose(method=Image.FLIP_TOP_BOTTOM).convert("RGB")
        else:
            self.texture_image = None
        self.texture_alpha = texture_alpha

        # Enable rendering passes
        self.cast_shadow = cast_shadow
        self.fragmap = True
        self.depth_prepass = True

        # Misc.
        self.flat_shading = False
        self.show_texture = self.has_texture
        self.norm_coloring = False
        self.normals_r = None
        self.need_upload = True

    @property
    def vertices(self):
        return self._vertices

    @vertices.setter
    def vertices(self, vertices):
        # Update vertices and redraw
        self._vertices = vertices
        self.n_frames = len(vertices)

        # If vertex normals were supplied, they are no longer valid.
        self._vertex_normals = None

        # Must clear all LRU caches where the vertices are used.
        self.compute_vertex_and_face_normals.cache_clear()

        self.redraw()

    @property
    def n_faces(self):
        return self.faces.shape[0]

    @property
    def n_vertices(self):
        return self.vertices.shape[1]

    @property
    def vertex_faces(self):
        # To compute the normals we need to know a mapping from vertex ID to all faces that this vertex is part of.
        # Because we are lazy we abuse trimesh to compute this for us. Not all vertices have the maximum degree, so
        # this array is padded with -1 if necessary.
        return trimesh.Trimesh(self.vertices[0], self.faces, process=False).vertex_faces

    @property
    def vertex_normals(self):
        """Get or compute all vertex normals (this might take a while for long sequences)."""
        if self._vertex_normals is None:
            vertex_normals, _ = compute_vertex_and_face_normals(self.vertices, self.faces, self.vertex_faces,
                                                                normalize=True)
            if vertex_normals.ndim < 3:
                vertex_normals = vertex_normals.unsqueeze(0)
            self._vertex_normals = vertex_normals
        return self._vertex_normals

    @property
    def face_normals(self):
        """Get or compute all face normals (this might take a while for long sequences)."""
        if self._face_normals is None:
            _, face_normals = compute_vertex_and_face_normals(self.vertices, self.faces, self.vertex_faces,
                                                              normalize=True)
            if face_normals.ndim < 3:
                face_normals = face_normals.unsqueeze(0)
            self._face_normals = face_normals
        return self._face_normals

    def vertex_normals_at(self, frame_id):
        """Get or compute the vertex normals at the given frame."""
        if self._vertex_normals is None:
            vn, _ = self.compute_vertex_and_face_normals(frame_id, normalize=True)
        else:
            assert len(self._vertex_normals.shape) == 3, f"Got shape {self._vertex_normals.shape}"
            vn = self._vertex_normals[frame_id]
        return vn

    def face_normals_at(self, frame_id):
        """Get or compute the face normals at the given frame."""
        if self._face_normals is None:
            _, fn = self.compute_vertex_and_face_normals(frame_id, normalize=True)
        else:
            assert len(self._face_normals.shape) == 3, f"Got shape {self._face_normals.shape}"
            fn = self._face_normals[frame_id]
        return fn

    @property
    def vertex_colors(self):
        return self._vertex_colors

    @vertex_colors.setter
    def vertex_colors(self, vertex_colors):
        # Vertex colors cannot be empty
        if vertex_colors is None:
            self._vertex_colors = np.full((self.n_frames, self.n_vertices, 4), self.color)
        elif isinstance(vertex_colors, tuple) and len(vertex_colors) == 4:
            self._vertex_colors = np.full((self.n_frames, self.n_vertices, 4), vertex_colors)
        else:
            self._vertex_colors = vertex_colors
        self.redraw()

    @property
    def face_colors(self):
        return self._face_colors

    @face_colors.setter
    def face_colors(self, face_colors):
        self._face_colors = face_colors
        if face_colors is not None:
            self._vertex_colors = np.tile(self.face_colors[:, :, np.newaxis], [1, 1, 3, 1])
        self.redraw()

    @Node.color.setter
    def color(self, color):
        only_alpha_changed = (np.abs((np.array(color) - np.array(self.color)))[-1] > 0
                              and (np.array(color)[:3] == np.array(self.color[:3])).all())

        self.material.color = color

        # If only alpha changed, don't update all colors
        if only_alpha_changed:
            self._vertex_colors[..., -1] = color[-1]
        # Otherwise, if no face colors, then override the existing vertex colors
        elif self.face_colors is None:
            self.vertex_colors = color

        # Update color buffer
        self.redraw()

    @property
    def current_vertices(self):
        return self.vertices[self.current_frame_id]

    @property
    def current_vertex_colors(self):
        return self.vertex_colors[self.current_frame_id]

    def closest_vertex_in_triangle(self, tri_id, point):
        face_vertex_id = np.linalg.norm((self.current_vertices[self.faces[tri_id]] - point), axis=-1).argmin()
        return self.faces[tri_id][face_vertex_id]

    def get_bc_coords_from_points(self, tri_id, points):
        return points_to_barycentric(self.current_vertices[self.faces[[tri_id]]], points)[0]

    @lru_cache(2048)
    def compute_vertex_and_face_normals(self, frame_id, normalize=False):
        """
        Compute face and vertex normals for the given frame. We use an LRU cache since this is a potentially
        expensive operation. This function exists because computing the normals on all frames can increase the
        startup time of the viewer considerably.

        :param frame_id: On which frame to compute the normals.
        :param normalize: Whether or not to normalize the normals. Not doing it is faster and the shaders typically
          enforce unit length of normals anyway.
        :return: The vertex and face normals as a np arrays of shape (V, 3) and (F, 3) respectively.
        """
        vs = self.vertices[frame_id:frame_id + 1]
        vn, fn = compute_vertex_and_face_normals(vs, self.faces, self.vertex_faces, normalize)
        return vn.squeeze(0), fn.squeeze(0)

    @property
    def bounds(self):
        return self.get_bounds(self.vertices)

    def on_frame_update(self):
        """Called whenever a new frame must be displayed."""
        super().on_frame_update()
        self.redraw()

    def _upload_buffers(self):
        """Upload the current frame data to the GPU for rendering."""
        if not self.is_renderable or not self._need_upload:
            return

        self._need_upload = False

        # Each write call takes about 1-2 ms
        vertices = self.current_vertices
        vertex_colors = self.current_vertex_colors

        if not self.flat_shading:
            vertex_normals = self.vertex_normals_at(self.current_frame_id)
            if self.face_colors is not None:
                vn_for_drawing = vertex_normals[self.faces]
            else:
                vn_for_drawing = vertex_normals
            self.vbo_normals.write(vn_for_drawing.astype('f4').tobytes())

        if self.face_colors is not None:
            vertices = vertices[self.faces]

        self.vbo_vertices.write(vertices.astype('f4').tobytes())
        self.vbo_colors.write(vertex_colors.astype('f4').tobytes())

        if self.has_texture:
            self.vbo_uvs.write(self.uv_coords.astype('f4').tobytes())

    def redraw(self):
        self._need_upload = True

    # noinspection PyAttributeOutsideInit
    @Node.once
    def make_renderable(self, ctx):
        """Prepares this object for rendering. This function must be called before `render` is used."""
        self.smooth_prog = get_smooth_lit_with_edges_program()
        self.flat_prog = get_flat_lit_with_edges_program()

        vertices = self.current_vertices
        vertex_normals = self.vertex_normals_at(self.current_frame_id)
        vertex_colors = self.current_vertex_colors

        # Face colors draw 3 distinct vertices per face.
        if self.face_colors is not None:
            vertices = vertices[self.faces]
            vn_for_drawing = vertex_normals[self.faces]
        else:
            vn_for_drawing = vertex_normals

        self.vbo_vertices = ctx.buffer(vertices.astype('f4').tobytes())
        self.vbo_normals = ctx.buffer(vn_for_drawing.astype('f4').tobytes())
        self.vbo_indices = ctx.buffer(self.faces.tobytes()) if self.face_colors is None else None
        self.vbo_colors = ctx.buffer(vertex_colors.astype('f4').tobytes())

        self.smooth_vao = ctx.vertex_array(self.smooth_prog,
                                           [(self.vbo_vertices, '3f4 /v', 'in_position'),
                                            (self.vbo_normals, '3f4 /v', 'in_normal'),
                                            (self.vbo_colors, '4f4 /v', 'in_color')],
                                           self.vbo_indices)

        self.flat_vao = ctx.vertex_array(self.flat_prog,
                                         [(self.vbo_vertices, '3f4 /v', 'in_position'),
                                          (self.vbo_colors, '4f4 /v', 'in_color')],
                                         self.vbo_indices)

        self.positions_vao = VAO('{}:positions'.format(self.unique_name))
        self.positions_vao.buffer(self.vbo_vertices, '3f', ['in_position'])
        self.positions_vao.index_buffer(self.vbo_indices)

        if self.has_texture:
            img = self.texture_image
            if self.use_pickle_texture:
                self.texture = ctx.texture(img.shape[:2], img.shape[2], img.tobytes())
            else:
                self.texture = ctx.texture(img.size, 3, img.tobytes())
            self.texture_prog = get_smooth_lit_texturized_program()
            self.vbo_uvs = ctx.buffer(self.uv_coords.astype('f4').tobytes())
            self.texture_vao = ctx.vertex_array(self.texture_prog,
                                                [(self.vbo_vertices, '3f4 /v', 'in_position'),
                                                 (self.vbo_normals, '3f4 /v', 'in_normal'),
                                                 (self.vbo_uvs, '2f4 /v', 'in_uv')],
                                                self.vbo_indices)

    @hooked
    def release(self):
        self.smooth_vao.release()
        self.flat_vao.release()
        self.positions_vao.release(buffer=False)

        self.vbo_vertices.release()
        self.vbo_indices.release()
        self.vbo_normals.release()
        self.vbo_colors.release()

        if self.has_texture:
            self.texture_vao.release()
            self.vbo_uvs.release()
            self.texture.release()

    def render(self, camera, **kwargs):
        # Check if flat shading changed, in which case we need to update the VBOs.
        flat = kwargs.get('flat_rendering', False)
        if flat != self.flat_shading:
            self.flat_shading = flat
            self.redraw()
            
        if self.has_texture:
            self.texture_prog['texture_alpha'].value = self.texture_alpha
        vao = self._prepare_vao(camera, **kwargs)
        vao.render(moderngl.TRIANGLES)

    def render_positions(self, prog):
        if self.is_renderable:
            self._upload_buffers()
            self.positions_vao.render(prog)

    def _prepare_vao(self, camera, **kwargs):
        """Prepare the shader pipeline and the VAO."""
        self._upload_buffers()

        if self.has_texture and self.show_texture:
            prog, vao = self.texture_prog, self.texture_vao
            prog['diffuse_texture'] = 0
            self.texture.use(0)
        else:
            prog, vao = (self.flat_prog, self.flat_vao) if self.flat_shading else (self.smooth_prog, self.smooth_vao)
            prog['norm_coloring'].value = self.norm_coloring

        prog['draw_edges'].value = 1.0 if kwargs['draw_edges'] and self.material._show_edges else 0.0
        prog['win_size'].value = kwargs['window_size']

        self.set_camera_matrices(prog, camera, **kwargs)
        set_lights_in_program(prog, kwargs['lights'], kwargs['shadows_enabled'])
        set_material_properties(prog, self.material)
        self.receive_shadow(prog, **kwargs)
        return vao

    def _show_normals(self):
        """Create and add normals at runtime"""
        vn = self.vertex_normals
        length = 0.01  # TODO make this either a parameter in the GUI or dependent on the bounding box.
        vn = vn / np.linalg.norm(vn, axis=-1, keepdims=True) * length

        # Must import here because if we do it at the top we create a circular dependency.
        from aitviewer.renderables.arrows import Arrows
        self.normals_r = Arrows(self.vertices, self.vertices + vn,
                                r_base=length / 10, r_head=2 * length / 10, p=0.25, name='Normals')
        self.normals_r.position = self.position
        self.normals_r.current_frame_id = self.current_frame_id
        self.add(self.normals_r, gui_elements=['material'])

    def gui(self, imgui):
        super(Meshes, self).gui(imgui)

        if self.has_texture:
            _, self.texture_alpha = imgui.slider_float('Texture alpha##texture_alpha{}'.format(self.unique_name),
                                                       self.texture_alpha, 0.0, 1.0, '%.2f')
        _, self.show_texture = imgui.checkbox('Render Texture##render_texture{}'.format(self.unique_name),
                                              self.show_texture)
        _, self.norm_coloring = imgui.checkbox('Norm Coloring##norm_coloring{}'.format(self.unique_name),
                                               self.norm_coloring)

        # TODO: Add  export workflow for all nodes
        if imgui.button('Export OBJ##export_{}'.format(self.unique_name)):
            mesh = trimesh.Trimesh(vertices=self.current_vertices, faces=self.faces, process=False)
            mesh.export('../export/' + self.name + '.obj')

        if self.normals_r is None:
            if imgui.button('Show Normals ##show_normals{}'.format(self.unique_name)):
                self._show_normals()


class VariableTopologyMeshes(Node):
    """
    A sequence of meshes that do not share the same topology (i.e. different nr of vertices, faces, texture, etc.).
    This simply treats every time instance as a separate mesh, so it's not optimized for performance.
    """

    def __init__(self,
                 vertices,
                 faces,
                 vertex_normals=None,
                 face_normals=None,
                 vertex_colors=None,
                 face_colors=None,
                 uv_coords=None,
                 texture_paths=None,
                 preload=True,
                 **kwargs):
        """
        Initializer.
        :param vertices: A list of length N with np arrays of size (V_n, 3).
        :param faces: A list of length N with np arrays of shape (F_n, 3).
        :param vertex_normals: An optional list of length N with np arrays of shape (V_n, 3).
        :param face_normals: An optional list of length N with np arrays of shape (F_n, 3).
        :param vertex_colors: An optional list of length N with np arrays of shape (V_n, 4) overriding the uniform color.
        :param face_colors: An optional list of length N with np arrays of shape (F_n, 4) overriding the uniform or vertex colors.
        :param uv_coords: An optional list of length N with np arrays of shape (V_n, 2) if the mesh is to be textured.
        :param texture_paths: An optional list of length N containing paths to the texture as an image file.
        :param preload: Whether or not to pre-load all the meshes. This increases loading time and memory consumption,
          but allows interactive animations.
        """
        assert len(vertices) == len(faces)
        super(VariableTopologyMeshes, self).__init__(n_frames=len(vertices))
        self.preload = preload

        self.vertices = vertices
        self.faces = faces
        self.vertex_normals = vertex_normals
        self.face_normals = face_normals
        self.vertex_colors = vertex_colors
        self.face_colors = face_colors
        self.uv_coords = uv_coords
        self.texture_paths = texture_paths
        self.mesh_kwargs = kwargs

        self._current_mesh = dict()  # maps from frame ID to mesh
        self._all_meshes = []
        if self.preload:
            for f in range(self.n_frames):
                m = self._construct_mesh_at_frame(f)
                self._all_meshes.append(m)
        # Set to true after changing the color and used when preloading 
        # is not enabled to override the current mesh color with the new color
        self._override_color = False

        self.show_texture = True
        self.norm_coloring = False
        self.ctx = None

    def _construct_mesh_at_frame(self, frame_id):
        m = Meshes(self.vertices[frame_id],
                   self.faces[frame_id],
                   self.vertex_normals[frame_id] if self.vertex_normals is not None else None,
                   self.face_normals[frame_id] if self.face_normals is not None else None,
                   self.vertex_colors[frame_id] if self.vertex_colors is not None else None,
                   self.face_colors[frame_id] if self.face_colors is not None else None,
                   self.uv_coords[frame_id] if self.uv_coords is not None else None,
                   self.texture_paths[frame_id] if self.texture_paths is not None else None,
                   **self.mesh_kwargs)
        return m

    @classmethod
    def from_trimeshes(cls, trimeshes, **kwargs):
        """Initialize from a list of trimeshes."""
        vertices = []
        faces = []
        vertex_normals = []
        for m in trimeshes:
            vertices.append(m.vertices)
            faces.append(m.faces)
            vertex_normals.append(m.vertex_normals)
        meshes = cls(vertices, faces, vertex_normals, **kwargs)
        return meshes

    @classmethod
    def from_plys(cls, plys, **kwargs):
        """Initialize from a list paths to .ply files."""
        vertices, faces, vertex_normals, vertex_colors = [], [], [], []
        sc = kwargs.get('vertex_scale', 1.0)
        for i in tqdm.tqdm(range(len(plys))):
            m = trimesh.load(plys[i])
            vertices.append(m.vertices * sc)
            faces.append(m.faces)
            vertex_normals.append(m.vertex_normals)
            vertex_colors.append(m.visual.vertex_colors / 255.0)
        return cls(vertices, faces, vertex_normals, vertex_colors=vertex_colors)

    @classmethod
    def from_directory(cls, path, preload=False, vertex_scale=1.0, high_quality=False, **kwargs):
        """ 
        Initialize from a directory containing mesh and texture data.

        Mesh files must be in pickle (.pkl) or obj (.obj) format, their name
        must start with 'mesh' and end with a frame number.

        Texture files must be in pickle (.pkl), png (.png) or jpeg (.jpg or .jpeg) format
        and their name must match the respective mesh filename with 'atlas' instead of 'mesh' 
        at the start of the filename.

        Example:
        path/mesh_001.obj
        path/mesh_002.obj
        path/atlas_001.png
        path/atlas_002.png

        A mesh pickle file must be a dictionary of numpy array with the following shapes:
        {
            'vertices': (V, 3)
            'normals':  (V, 3)
            'uvs':      (V, 2)
            'faces':    (F, 3)
        }

        A texture pickle file must be a raw RGB image stored as a numpy array of shape (width, height, 3)
        """

        files = os.listdir(path)

        # Supported mesh formats in order of preference (fastest to slowest)
        mesh_supported_formats = [".pkl", ".obj"]

        mesh_format = None
        for format in mesh_supported_formats:
            if any(map(lambda x: x.startswith("mesh") and x.endswith(format), files)):
                mesh_format = format
                break

        if mesh_format is None:
            raise ValueError(
                f'Unable to find mesh with supported extensions ({", ".join(mesh_supported_formats)}) at {path}')

        # Supported texture formats in order of preference (fastest to slowest)
        texture_supported_formats = [".pkl", ".jpg", ".jpeg", ".png"]

        # If high_quality is set to true prioritize PNGs
        if high_quality:
            texture_supported_formats = [".png", ".jpg", ".jpeg", ".pkl"]

        texture_format = None
        for format in texture_supported_formats:
            if any(map(lambda x: x.startswith("atlas") and x.endswith(format), files)):
                texture_format = format
                break

        if texture_format is None:
            raise ValueError(
                f'Unable to find atlas with supported extensions ({", ".join(texture_supported_formats)}) at {path}')

        # Load all objects sorted by the keyframe number specified in the file name
        regex = re.compile(r"(\d*)$")

        def sort_key(x):
            name = os.path.splitext(x)[0]
            return int(regex.search(name).group(0))

        obj_names = filter(lambda x: x.startswith("mesh") and x.endswith(mesh_format), files)
        obj_names = sorted(obj_names, key=sort_key)

        vertices, faces, vertex_normals, uvs = [], [], [], []
        texture_paths = []

        for obj_name in tqdm.tqdm(obj_names):
            if mesh_format == ".pkl":
                mesh = pickle.load(open(os.path.join(path, obj_name), "rb"))
                vertices.append(mesh['vertices'] * vertex_scale)
                faces.append(mesh['faces'])
                vertex_normals.append(mesh['normals'])
                uvs.append(mesh['uvs'])
            else:
                mesh = trimesh.load(os.path.join(path, obj_name), process=False)
                vertices.append(mesh.vertices * vertex_scale)
                faces.append(mesh.faces)
                vertex_normals.append(mesh.vertex_normals)
                uvs.append(np.array(mesh.visual.uv).squeeze())

            texture_paths.append(
                os.path.join(path, obj_name.replace("mesh", "atlas").replace(mesh_format, texture_format)))

        return cls(vertices, faces, vertex_normals, uv_coords=uvs,
                   texture_paths=texture_paths, preload=preload, **kwargs)

    @property
    def current_mesh(self):
        if self.preload: 
            return self._all_meshes[self.current_frame_id]
        else:
            m = self._current_mesh.get(self.current_frame_id, None)
            if m is None:
                # Need to construct a new one and clean up the old one.
                for k in self._current_mesh:
                    self._current_mesh[k].release()
                m = self._construct_mesh_at_frame(self.current_frame_id)

                # Set mesh position and scale
                m.position = self.position
                m.scale = self.scale

                # Set mesh material
                m.material = self.material

                # Override the mesh color only if it has been changed by the user
                if self._override_color:
                    m.color = self.color

                m.make_renderable(self.ctx)
                self._current_mesh = {self.current_frame_id: m}
            return m

    @property
    def bounds(self):
        return self.current_mesh.get_bounds(self.current_mesh.vertices)

    # noinspection PyAttributeOutsideInit
    @Node.once
    def make_renderable(self, ctx):
        if self.preload:
            for m in self._all_meshes:
                m.make_renderable(ctx)
        else:
            # A bit hacky so that we can make dynamically loaded meshes renderable when we load them.
            self.ctx = ctx
            self.current_mesh.make_renderable(ctx)

    def render(self, camera, **kwargs):
        self.current_mesh.show_texture = self.show_texture
        self.current_mesh.norm_coloring = self.norm_coloring
        self.current_mesh.render(camera, **kwargs)

    def render_depth_prepass(self, camera, **kwargs):
        self.current_mesh.render_depth_prepass(camera, **kwargs)

    def render_shadowmap(self, light_mvp, program):
        self.current_mesh.render_shadowmap(light_mvp, program)

    def gui_position(self, imgui):
        # Position controls
        u, pos = imgui.drag_float3('Position##pos{}'.format(self.unique_name), *self.position, 0.1, format='%.2f')
        if u:
            self.position = pos
            # If meshes are already loaded go through all and update the positions
            if self.preload:
                for m in self._all_meshes:
                    m.position = pos
            # Otherwise only update the current mesh, other meshes will be updated when loaded
            else:
                self.current_mesh.position = pos

    def gui_scale(self, imgui):
        # Scale controls
        u, scale = imgui.drag_float('Scale##scale{}'.format(self.unique_name), self.scale, 0.01, min_value=0.001,
                                    max_value=10.0, format='%.3f')
        if u:
            self.scale = scale
            # If meshes are already loaded go through all and update the scale factors
            if self.preload:
                for m in self._all_meshes:
                    m.scale = scale
            # Otherwise only update the current mesh, other meshes will be updated when loaded
            else:
                self.current_mesh.scale = scale

    def gui_material(self, imgui, show_advanced=True):
        # Color Control
        uc, color = imgui.color_edit4("Color##color{}'".format(self.unique_name), *self.material.color, show_alpha=True)
        if uc:
            self.color = color
            # If meshes are already loaded go through all and update the vertex colors
            if self.preload:
                for m in self._all_meshes:
                    m.color = color
            # Otherwise only update the current mesh and enable color override for other meshes
            else:
                self.current_mesh.color = color
                self._override_color = True

        _, self.show_texture = imgui.checkbox('Render Texture', self.show_texture)
        _, self.norm_coloring = imgui.checkbox('Norm Coloring', self.norm_coloring)

        if show_advanced:
            if imgui.tree_node("Advanced material##advanced_material{}'".format(self.unique_name)):
                # Diffuse
                ud, diffuse = imgui.slider_float('Diffuse##diffuse{}'.format(self.unique_name),
                                                 self.current_mesh.material.diffuse,
                                                 0.0, 1.0, '%.2f')

                if ud:
                    self.material.diffuse = diffuse
                    # If meshes are already loaded go through all and update the diffuse value
                    if self.preload:
                        for m in self._all_meshes:
                            m.material.diffuse = diffuse
                    # Otherwise only update the current mesh, other meshes will be updated when loaded
                    else:
                        self.current_mesh.material.diffuse = diffuse

                # Ambient
                ua, ambient = imgui.slider_float('Ambient##ambient{}'.format(self.unique_name),
                                                 self.current_mesh.material.ambient,
                                                 0.0, 1.0, '%.2f')
                if ua:
                    self.material.ambient = ambient
                    # If meshes are already loaded go through all and update the ambient value
                    if self.preload:
                        for m in self._all_meshes:
                            m.material.ambient = ambient
                    # Otherwise only update the current mesh, other meshes will be updated when loaded
                    else:
                        self.current_mesh.material.ambient = ambient

                imgui.tree_pop()



"""
Copyright (C) 2022  ETH Zurich, Manuel Kaufmann, Velko Vechev

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import moderngl
import numpy as np

from aitviewer.renderables.coordinate_system import CoordinateSystem
from aitviewer.renderables.meshes import Meshes
from aitviewer.renderables.plane import ChessboardPlane
from aitviewer.scene.light import Light
from aitviewer.scene.node import Node


class Scene(Node):
    """Generic scene node"""

    def __init__(self, **kwargs):
        """Create a scene with a name."""
        super(Scene, self).__init__(**kwargs)

        # References resources in the scene
        self.lights = []
        self.camera = None

        # Scene has a reference to ctx
        self.ctx = None

        self.backface_culling = True

        # Default Setup
        # If you update the number of lights, make sure to change the respective `define` statement in
        # directional_lights.glsl as well!
        # Influence of diffuse lighting is controlled globally for now, but should eventually be a material property.
        self.lights.append(Light(name='Back Light',  position=(0.0, 10.0, 15.0),  color=(1.0, 1.0, 1.0, 1.0)))
        self.lights.append(Light(name='Front Light', position=(0.0, 10.0, -15.0), color=(1.0, 1.0, 1.0, 1.0), shadow_enabled=False))
        self.add(*self.lights, show_in_hierarchy=False)

        # Scene items
        self.origin = CoordinateSystem(name="Origin", length=0.1)
        
        self.floor = ChessboardPlane(100.0, 200, (0.9, 0.9, 0.9, 1.0),  (0.82, 0.82, 0.82, 1.0), name="Floor")
        self.floor.material.diffuse = 0.1

        self.add(self.origin, has_gui=False)
        self.add(self.floor)

        self.custom_font = None

    def render(self, **kwargs):
        # As per https://learnopengl.com/Advanced-OpenGL/Blending

        # Collect all renderable nodes
        rs = self.collect_nodes()

        transparent = []

        # Draw all opaque objects first
        for r in rs:
            # An object is opaque if it's color alpha is 1.0 
            # and, if it's a mesh, if also the texture_alpha is 1.0
            if r.color[3] == 1.0 and (not isinstance(r, Meshes) or r.texture_alpha == 1.0):
                # Turn off backface culling if enabled for the scene
                # and requested by the current object
                if self.backface_culling and r.backface_culling:
                    self.ctx.enable(moderngl.CULL_FACE)
                else:
                    self.ctx.disable(moderngl.CULL_FACE)

                self.safe_render(r, **kwargs)
            else:
                # Otherwise append to transparent list
                transparent.append(r)
        
        # Draw back to front by sorting transparent objects based on distance to camera.
        # As an approximation we only sort using the origin of the object 
        # which may be incorrect for large objects or objects significantly 
        # offset from their origin, but works in many cases.
        for r in sorted(transparent, key=lambda r: np.linalg.norm(r.position - self.camera.position), reverse=True):
            # Render to depth buffer only
            self.ctx.depth_func = '<'
            self.safe_render_depth_prepass(r, **kwargs)

            # Turn off backface culling if enabled for the scene
            # and requested by the current object
            if self.backface_culling and r.backface_culling:
                self.ctx.enable(moderngl.CULL_FACE)
            else:
                self.ctx.disable(moderngl.CULL_FACE)

            # Render normally with less equal depth comparison function,
            # drawing only the pixels closer to the camera to avoid
            # order dependent blending artifacts
            self.ctx.depth_func = '<='
            self.safe_render(r, **kwargs)

        # Restore the default depth comparison function
        self.ctx.depth_func = '<'
    
    def safe_render_depth_prepass(self, r, **kwargs):
        if not r.is_renderable:
            r.make_renderable(self.ctx)
        r.render_depth_prepass(self.camera, **kwargs)

    def safe_render(self, r, **kwargs):
        if not r.is_renderable:
            r.make_renderable(self.ctx)
        r.render(self.camera, **kwargs)

    def make_renderable(self, ctx):
        self.ctx = ctx
        rs = self.collect_nodes(req_enabled=False)
        for r in rs:
            r.make_renderable(self.ctx)

    def auto_set_floor(self):
        """Finds the minimum lower bound in the y coordinate from all the children bounds and uses that as the floor"""
        rs = self.collect_nodes()
        collected_bounds = []
        for r in rs:
            if r.bounds is not None:
                collected_bounds.append(r.bounds)

        if len(collected_bounds) > 0:
            self.floor.position[1] = np.array(collected_bounds)[:, 1, 0].min()

    def auto_set_camera_target(self):
        """Sets the camera target to the average of the center of all objects in the scene"""
        rs = self.collect_nodes()
        centers = []
        for r in rs:
            if r.bounds is not None:
                center = np.matmul(r.rotation, r.bounds.mean(-1)) + r.position
                if center.sum() != 0.0:
                    centers.append(center)

        if len(centers) > 0:
            self.camera.target = np.array(centers).mean(0)

    def set_lights(self, is_dark_mode=False):
        if is_dark_mode:
            for l in self.lights:
                l.intensity_ambient = 0.2
        else:
            for l in self.lights:
                l.intensity_ambient = 1.0

    def collect_nodes(self, req_enabled=True, obj_type=Node):
        nodes = []

        # Use head recursion in order to collect the most deep nodes first
        # These are likely to be nodes which should be rendered first (i.e. transparent parent nodes should be last)
        def rec_collect_nodes(n):
            if (not req_enabled or n.enabled):
                if isinstance(n, obj_type):
                    nodes.append(n)
                for n_child in n.nodes:
                    rec_collect_nodes(n_child)

        for n in self.nodes:
            rec_collect_nodes(n)
        return nodes

    def get_node_by_name(self, name):
        assert name != ''
        ns = self.collect_nodes()
        for n in ns:
            if n.name == name:
                return n
        return None

    def get_node_by_uid(self, uid):
        ns = self.collect_nodes()
        for n in ns:
            if n.uid == uid:
                return n
        return None

    def gui(self, imgui):
        """GUI to control scene settings."""
        self.gui_camera(imgui)
        self.gui_lights(imgui)
        self.gui_renderables(imgui, self.nodes)
        
    def gui_camera(self, imgui):
        # Camera GUI
        imgui.push_font(self.custom_font)
        camera_expanded = imgui.tree_node("\u0084 Camera##tree_node_r_camera")
        imgui.pop_font()
        if camera_expanded:
            self.camera.gui(imgui)
            imgui.tree_pop()
        imgui.spacing()

    def gui_lights(self, imgui):
        # Lights GUI
        for _, light in enumerate(self.lights):
            imgui.push_font(self.custom_font)
            light_expanded = imgui.tree_node("\u0085 {}##tree_node_r".format(light.name))
            imgui.pop_font()
            if light_expanded:
                light.gui(imgui)
                imgui.tree_pop()
                imgui.spacing()

    def gui_renderables(self, imgui, rs):
        # Nodes GUI
        for i, r in enumerate(rs):
            if r.show_in_hierarchy:
                # Visibility
                curr_enabled = r.enabled
                if not curr_enabled:
                    imgui.push_style_color(imgui.COLOR_TEXT, 1.0, 1.0, 1.0, 0.4)

                # Title
                imgui.push_font(self.custom_font)
                if r.has_gui:
                    r.expanded = imgui.tree_node("{} {}##tree_node_{}".format(r.icon, r.name, r.unique_name),
                                                 r.expanded and imgui.TREE_NODE_DEFAULT_OPEN)
                else:
                    imgui.spacing()
                    imgui.same_line(spacing=24)
                    imgui.text("{} {}".format(r.icon, r.name))
                imgui.pop_font()

                # Aligns checkbox to the right side of the window
                # https://github.com/ocornut/imgui/issues/196
                imgui.same_line(position=imgui.get_window_content_region_max().x - 25)
                eu, enabled = imgui.checkbox('##enabled_r_{}'.format(r.unique_name), r.enabled)
                if eu:
                    r.enabled = enabled

                if r.has_gui and r.expanded:
                    # Render this elements GUI
                    r.gui(imgui)
                    # Recursively render the GUI of this elements children renderables
                    self.gui_renderables(imgui, r.nodes)

                    imgui.spacing()
                    imgui.tree_pop()

                if not curr_enabled:
                    imgui.pop_style_color(1)

    def add_light(self, light):
        self.lights.append(light)

    @property
    def n_lights(self):
        return len(self.lights)

    @property
    def n_frames(self):
        n_frames = 1
        ns = self.collect_nodes(req_enabled=False)
        for n in ns:
            n_frames = max(n_frames, n.n_frames)
        return n_frames



"""
Copyright (C) 2022  ETH Zurich, Manuel Kaufmann, Velko Vechev

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import numpy as np

from aitviewer.scene.node import Node
from aitviewer.renderables.rigid_bodies import RigidBodies


class CoordinateSystem(Node):
    """
    Render a coordinate system using shaded cylinders.
    """

    def __init__(self,
                 length=1.0,
                 icon="\u0086",
                 **kwargs):
        super(CoordinateSystem, self).__init__(icon=icon, **kwargs)

        r = length / 50
        l = length

        self.rb = RigidBodies(np.array([[[0.0, 0.0, 0.0]]]), np.eye(3)[np.newaxis, np.newaxis], radius=r, length=l)
        self.add(self.rb)
